#include "pacer.h"
#include "timeutil.h"

#include <algorithm>

using namespace std;

Pacer::Pacer(Sender *sender, int max_rate_kbit_per_sec, int burst_num_packets)
	: sender(sender), burst_num_packets(burst_num_packets), next_send_packet{0, 0}
{
	const int max_rate_byte_per_sec = 1000 * max_rate_kbit_per_sec / 8;
	seconds_per_byte = 1.0 / max_rate_byte_per_sec;
}

void Pacer::send_packet(uint16_t proto, const string &data, uint32_t incoming_seq)
{
	waiting_packets.push_back(GREPacket{incoming_seq, proto, data, {0, 0}});
	possibly_flush_packets();
}

void Pacer::possibly_adjust_tv(timeval *tv)
{
	if (waiting_packets.empty()) {
		return;
	}

	timeval now;
	gettimeofday(&now, NULL);

	timeval tdiff = subtract_timeval_saturate(next_send_packet, now);
	if (less_than(tdiff, *tv)) {
		*tv = tdiff;
	}
}

void Pacer::possibly_flush_packets()
{
	if (waiting_packets.empty()) {
		return;
	}

	timeval now;
	gettimeofday(&now, NULL);
	if (less_than(now, next_send_packet)) {
		return;
	}

	int bytes_sent = 0;
	for (int i = 0; i < burst_num_packets && !waiting_packets.empty(); ++i) {
		const GREPacket &packet = waiting_packets.front();
		sender->send_packet(packet.proto, packet.data, packet.seq);
		bytes_sent += packet.data.size();
		waiting_packets.pop_front();
	}

	next_send_packet = offset_timeval_seconds(now, seconds_per_byte * bytes_sent);
}

