#include <string.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <fecpp.h>

#include "reorderer.h"
#include "rsencoder.h"
#include "rs_parm.h"

#include <algorithm>

using namespace std;
	
RSEncoder::RSEncoder(Sender *sender) 
	: sender(sender),
	  rs(RS_PAYLOAD_SIZE, RS_GROUP_SIZE)
{
}

void RSEncoder::send_packet(uint16_t proto, const std::string &data, uint32_t incoming_seq)
{
	if (!packet_history.empty() &&
	    incoming_seq <= packet_history.back().seq) {
		// Reorderer should have done this for us.
		return;
	}
	if (!packet_history.empty() &&
	    incoming_seq / RS_PAYLOAD_SIZE !=
		packet_history.back().seq / RS_PAYLOAD_SIZE) {
		// Received an unfinished group.
		packet_history.clear();
	}
	bool debug_drop_packet = false;  // For testing only.
	if (incoming_seq % RS_PAYLOAD_SIZE == 3) {
		//debug_drop_packet = true;
	}
	if (debug_drop_packet) {
		const unsigned char *ptr = reinterpret_cast<const unsigned char *>(data.data());
		printf("DEBUG: Dropping packet seq=%u proto=0x%04x len=%d data=%02x %02x %02x %02x %02x %02x %02x %02x ...\n",
			incoming_seq, proto, data.size(), ptr[0], ptr[1], ptr[2], ptr[3], ptr[4], ptr[5], ptr[6], ptr[7]);
	} else {
		sender->send_packet(proto, data, incoming_seq);
	}
	packet_history.emplace_back(GREPacket{incoming_seq, proto, data});
	if (packet_history.size() == RS_PAYLOAD_SIZE) {
		finish_group();
	}
}

void RSEncoder::finish_group()
{
	// Our RS packets need to have the same max length as the longest one.
	int max_length = 0;
	for (int i = 0; i < packet_history.size(); ++i) {
		max_length = max<int>(max_length, packet_history[i].data.size());
	}

	string padded_packets;
	padded_packets.reserve((max_length + 4) * packet_history.size());
	// TODO: RS_PAD
	for (int i = 0; i < packet_history.size(); ++i) {
		string p;
		p.resize(max_length + 4);
		memset(&p[0], 0, max_length + 4);
		uint16_t proto_be = htons(packet_history[i].proto);
		memcpy(&p[0], &proto_be, sizeof(uint16_t));
		uint16_t len_be = htons(packet_history[i].data.size());
		memcpy(&p[2], &len_be, sizeof(uint16_t));
		memcpy(&p[4], packet_history[i].data.data(), packet_history[i].data.size());
		padded_packets += p;
	}

	// Now construct and send RS packets.
	rs.encode(reinterpret_cast<const fecpp::byte*>(padded_packets.data()),
	          padded_packets.size(),
	          [&](size_t packet_num, size_t num_packets, const fecpp::byte data[], size_t size) {
			// The first N packets are just the original ones; ignore them.
			if (packet_num < RS_PAYLOAD_SIZE) {
				return;
			}

			const char *sdata = reinterpret_cast<const char *>(data);
			uint32_t start_seq = packet_history[0].seq - 1;
			sender->send_packet(0xffff, string(sdata, size), start_seq - (packet_num - RS_PAYLOAD_SIZE));
		});
	
	packet_history.clear();
}

